import os
import sys
import warnings

# Check if JuliaCall is already loaded, and if so, warn the user
# about the relevant environment variables. If not loaded,
# set up sensible defaults.
if "juliacall" in sys.modules:
    warnings.warn(
        "juliacall module already imported. "
        "Make sure that you have set the environment variable `PYTHON_JULIACALL_HANDLE_SIGNALS=yes` to avoid segfaults. "
        "Also note that OCTOFITTERPY will not be able to configure `PYTHON_JULIACALL_THREADS` or `PYTHON_JULIACALL_OPTLEVEL` for you."
    )
else:
    # Required to avoid segfaults (https://juliapy.github.io/PythonCall.jl/dev/faq/)
    if os.environ.get("PYTHON_JULIACALL_HANDLE_SIGNALS", "yes") != "yes":
        warnings.warn(
            "PYTHON_JULIACALL_HANDLE_SIGNALS environment variable is set to something other than 'yes' or ''. "
            + "You will experience segfaults if running with multithreading."
        )

    if os.environ.get("PYTHON_JULIACALL_THREADS", "auto") != "auto":
        warnings.warn(
            "PYTHON_JULIACALL_THREADS environment variable is set to something other than 'auto', "
            "so OCTOFITTERPY was not able to set it. You may wish to set it to `'auto'` for full use "
            "of your CPU."
        )

    # TODO: Remove these when juliapkg lets you specify this
    for k, default in (
        ("PYTHON_JULIACALL_HANDLE_SIGNALS", "yes"),
        ("PYTHON_JULIACALL_THREADS", "auto"),
        ("PYTHON_JULIACALL_OPTLEVEL", "3"),
    ):
        os.environ[k] = os.environ.get(k, default)


from juliacall import Main as jl  # type: ignore

# Overwrite the seval function to use Meta.parseall
# instead of Meta.parse.
jl.seval("using PythonCall: PythonCall, Py, pyconvert")
jl.seval(
    """function PythonCall.pyjlmodule_seval(self::Module, expr::Py)
    e = Meta.parseall(strip(pyconvert(String, expr)))
    Py(Base.eval(self, e))
end"""
)
# ^TODO: Overwrite this once PythonCall.jl is updated:

jl_version = (jl.VERSION.major, jl.VERSION.minor, jl.VERSION.patch)

# Next, automatically load the juliacall extension if we're in a Jupyter notebook
autoload_extensions = os.environ.get("OCTOFITTERPY_AUTOLOAD_EXTENSIONS", "yes")
if autoload_extensions in {"yes", ""} and jl_version >= (1, 9, 0):
    try:
        get_ipython = sys.modules["IPython"].get_ipython

        if "IPKernelApp" not in get_ipython().config:
            raise ImportError("console")

        print(
            "Detected Jupyter notebook. Loading juliacall extension. Set `OCTOFITTERPY_AUTOLOAD_EXTENSIONS=no` to disable."
        )

        # TODO: Turn this off if juliacall does this automatically
        get_ipython().run_line_magic("load_ext", "juliacall")
    except Exception:
        pass
elif autoload_extensions not in {"no", "yes", ""}:
    warnings.warn(
        "OCTOFITTERPY_AUTOLOAD_EXTENSIONS environment variable is set to something other than 'yes' or 'no' or ''."
    )

jl.seval("using Octofitter")
Octofitter = jl.Octofitter



from octofitterpy import Octofitter
from octofitterpy import jl
from .julia_helpers import jl_array
import sys

# Inside Planet and System blocks, which are resolved in the Main module,
# we need various names available to the user
jl.seval("using Octofitter, Distributions")

gaia_plx = Octofitter.gaia_plx
HGCALikelihood = Octofitter.HGCALikelihood
LogDensityModel = Octofitter.LogDensityModel
octofit = Octofitter.octofit
loadchain = Octofitter.loadchain
savechain = Octofitter.savechain
mjd = Octofitter.mjd
mjd2date = Octofitter.mjd2date
years2mjd = Octofitter.mjd2date

def Planet(
    name,
    basis,
    priors,
    likelihoods=[]
):
    try:
        len(likelihoods)
    except:
        raise ValueError("likelihoods must be a list of likelihood objects (even if you are only providing one value)")
    jl.Main._obs = tuple(likelihoods)
    expr = f"""
        @planet {name} {basis} begin
            {priors}
        end (_obs...)
    """
    planet = jl.seval(expr)
    return planet

def System(
    name,
    priors,
    likelihoods=[],
    companions=[],
):
    try:
        len(likelihoods)
    except:
        raise ValueError("likelihoods must be a list of likelihood objects (even if you are only providing one value)")
    try:
        len(companions)
    except:
        raise ValueError("companions must be a list of Planet objects (even if you are only providing one value)")
    jl.Main._obs = tuple(likelihoods)
    jl.Main._plnt = tuple(companions)
    expr = f"""
        @system {name} begin
            {priors}
        end (_plnt...)
    """
    # _plnt
    sys = jl.seval(expr)
    return sys

def PlanetRelAstromLikelihood(**data):
    for k,v in data.items():
        data[k] = jl_array(v)
    return Octofitter.PlanetRelAstromLikelihood(Octofitter.Table(**data))


# These functions require us to load a plotting backend, which is a little
# slow. Only load it when we need it.
def octoplot(*args, **kwargs):
    jl.seval("using Plots: Plots")
    fig = Octofitter.octoplot(*args, **kwargs)
    if isipynb():
        fname = jl.tempname()+".png"
        jl.Main.Plots.savefig(fig, fname)
        from IPython.display import Image
        return Image(filename=fname) 


def octocorner(*args, **kwargs):
    jl.seval("using CairoMakie: Makie")
    jl.seval("using PairPlots")
    fig = Octofitter.octocorner(*args, **kwargs)
    if isipynb():
        fname = jl.tempname()+".png"
        jl.Main.Makie.save(fname, fig)
        from IPython.display import Image
        return Image(filename=fname) 

def isipynb():
    try:
        get_ipython = sys.modules["IPython"].get_ipython

        if "IPKernelApp" not in get_ipython().config:
            raise ImportError("console")

        return True
    except Exception:
        return False


import octofitterpy as octo

astrom_like = octo.PlanetRelAstromLikelihood(
    # MJD
    epoch = [50000,50120,50240,50360,50480,50600,50720,50840],
    # delta RA, milliarcseconds (East is positive)
    ra = [-505.7,-502.5,-498.2,-492.6,-485.9,-478.1,-469.0,-458.8],
    # delta DEC, milliarcseconds (North is positive)
    dec = [-66.9,-37.4,-7.9,21.6,51.1,80.5,109.7,138.6],
    # Uncertainty on RA, milliarcseconds
    σ_ra = [10,10,10,10,10,10,10,10.0],
    # Uncertainty on DEC, milliarcseconds
	σ_dec = [10,10,10,10,10,10,10,10],
    # Corelation between RA and DEC uncertainties
	cor= [0,0,0,0,0,0,0,0.0]
)

# You can also use separation (milliarcseconds) and position angle (RADIANS):
# astrom_like = octo.PlanetRelAstromLikelihood(
#     # MJD
#     epoch = [50000,50120],
#     # separation from star, milliarcseconds (East is positive)
#     sep = [505.7,600.1],
#     # position angle, RADIANS (!!) (NOT DEGREES)
#     pa = [0.0,0.4,],
#     # Uncertainty on sep, milliarcseconds
#     σ_sep = [10,10],
#     # Uncertainty on pa, RADIANS
# 	  σ_pa = [0.01,0.01],
#     # Corelation between PA and SEP uncertainties
#     cor= [0,0.2]
# )

# To convert dates:
#  *  octo.mjd("2020-01-01") -> to MJD
#  *  octo.years2mjd("2020.0") -> to MJD
#  *  octo.mjd2date(50000) -> datetime.date(1995, 10, 10)

b = octo.Planet(
    name="b",
    basis="Visual{KepOrbit}",
    priors=
    """            
        a ~ LogUniform(0.1, 500)
        e ~ Uniform(0.0, 0.99)
        i ~ Sine()
        ω ~ UniformCircular()
        Ω ~ UniformCircular()
        θ ~ UniformCircular()
        tp = θ_at_epoch_to_tperi(system,b,50000) # use MJD epoch of your data here!!
    """,
    likelihoods=[astrom_like]
)

# The name of your system determines the output file names
sys = octo.System(
    name="HIP100123",
    priors = 
    """
        M ~ truncated(Normal(1.2, 0.1), lower=0)
        plx ~ truncated(Normal(50.0, 0.02), lower=0)
    """,
    likelihoods=[],
    companions=[b]
)

model = octo.LogDensityModel(sys)

# defaults: adaptation=1000, iterations=1000. Increase both for more points.
chain = octo.octofit(model)

## Summary
summary = repr(chain)
print(summary)

## Plotting
octo.octoplot(model,chain)
octo.octocorner(model,chain,small=True)

## Saving (use octo.loadchain to restore)
octo.savechain(chain)


"""Functions for initializing the Julia environment and installing deps."""

import numpy as np
from juliacall import convert as jl_convert  # type: ignore

from .julia_import import jl

jl.seval("using PythonCall: PythonCall")

PythonCall = jl.PythonCall


# def _escape_filename(filename):
#     """Turn a path into a string with correctly escaped backslashes."""
#     str_repr = str(filename)
#     str_repr = str_repr.replace("\\", "\\\\")
#     return str_repr



def jl_array(x):
    if x is None:
        return None
    return jl_convert(jl.Array, x)


# def jl_serialize(obj):
#     buf = jl.IOBuffer()
#     Serialization.serialize(buf, obj)
#     return np.array(jl.take_b(buf))


# def jl_deserialize(s):
#     if s is None:
#         return s
#     buf = jl.IOBuffer()
#     jl.write(buf, jl_array(s))
#     jl.seekstart(buf)
#     return Serialization.deserialize(buf)



# This must be imported as early as possible to prevent
# library linking issues caused by numpy/pytorch/etc. importing
# old libraries:
from .julia_import import jl, Octofitter  # isort:skip

from .julia_helpers import jl_array
from .python_interface import *

# This file is created by setuptools_scm during the build process:
from .version import __version__

__all__ = [
    "jl",
    "Octofitter",
    "__version__",
]
